"""This module implements the Scraper component which parses responses and
extracts information from them"""
import asyncio
import logging
from collections import deque

from itemadapter import is_item
from scrapy import signals
from scrapy.exceptions import CloseSpider, DropItem, IgnoreRequest
from scrapy.http import Request, Response
from scrapy.utils.log import logformatter_adapter
from scrapy.utils.misc import load_object, warn_on_generator_with_return_value

from aioscrapy.middleware import SpiderMiddlewareManager
from aioscrapy.utils.tools import call_helper

logger = logging.getLogger(__name__)


class Slot:
    """Scraper slot (one per running spider)"""

    MIN_RESPONSE_SIZE = 1024

    def __init__(self, max_active_size=5000000):
        self.max_active_size = max_active_size
        self.queue = deque()
        self.active = set()
        self.active_size = 0
        self.itemproc_size = 0
        self.closing = None

    def add_response_request(self, response, request):
        self.queue.append((response, request))
        if isinstance(response, Response):
            self.active_size += max(len(response.body), self.MIN_RESPONSE_SIZE)
        else:
            self.active_size += self.MIN_RESPONSE_SIZE

    def next_response_request_deferred(self):
        response, request = self.queue.popleft()
        self.active.add(request)
        return response, request

    def finish_response(self, response, request):
        self.active.remove(request)
        if isinstance(response, Response):
            self.active_size -= max(len(response.body), self.MIN_RESPONSE_SIZE)
        else:
            self.active_size -= self.MIN_RESPONSE_SIZE

    def is_idle(self):
        return not (self.queue or self.active)

    def needs_backout(self):
        return self.active_size > self.max_active_size


class Scraper:

    def __init__(self, crawler):
        self.slot = None
        self.spidermw = SpiderMiddlewareManager.from_crawler(crawler)
        itemproc_cls = load_object(crawler.settings['ITEM_PROCESSOR'])
        self.itemproc = itemproc_cls.from_crawler(crawler)
        self.crawler = crawler
        self.signals = crawler.signals
        self.logformatter = crawler.logformatter
        self.concurrent_items_semaphore = asyncio.Semaphore(crawler.settings.getint('CONCURRENT_ITEMS'))

    async def open_spider(self, spider):
        """Open the given spider for scraping and allocate resources for it"""
        self.slot = Slot(self.crawler.settings.getint('SCRAPER_SLOT_MAX_ACTIVE_SIZE'))
        await self.itemproc.open_spider(spider)

    async def close_spider(self, spider):
        """Close a spider being scraped and release its resources"""
        slot = self.slot
        slot.closing = asyncio.Future()
        await self.itemproc.close_spider(spider)
        self._check_if_closing(spider, slot)
        await slot.closing

    def is_idle(self):
        """Return True if there isn't any more spiders to process"""
        return not self.slot

    def _check_if_closing(self, spider, slot):
        if slot.closing and slot.is_idle():
            slot.closing.set_result(spider)

    async def enqueue_scrape(self, response, request, spider):
        slot = self.slot
        # 将结果缓存到slot中
        slot.add_response_request(response, request)
        try:
            # 处理响应结果
            await self._scrape_next(spider, slot)
        except (Exception, BaseException) as e:
            logger.error('Scraper bug processing %(request)s',
                         {'request': request},
                         exc_info=response,
                         extra={'spider': spider})
        finally:
            # 将slot中的缓存结果删除
            slot.finish_response(response, request)
            # 检测slot还有未处理的任务
            self._check_if_closing(spider, slot)
            # 继续处理slot缓存中的任务
            asyncio.create_task(self._scrape_next(spider, slot))

    async def _scrape_next(self, spider, slot):
        while slot.queue:
            response, request = slot.next_response_request_deferred()
            await self._scrape(response, request, spider)

    async def _scrape(self, result, request, spider):
        """
        Handle the downloaded response or failure through the spider callback/errback
        """
        if not isinstance(result, (Response, Exception, BaseException)):
            raise TypeError(f"Incorrect type: expected Response or Failure, got {type(result)}: {result!r}")
        try:
            response = await self._scrape2(result, request, spider)  # returns spider's processed output
        except (Exception, BaseException) as e:
            await self.handle_spider_error(e, request, result, spider)
        else:
            await self.handle_spider_output(response, request, result, spider)

    async def _scrape2(self, result, request, spider):
        """
        Handle the different cases of request's result been a Response or a Failure
        """
        if isinstance(result, Response):
            # 将Response丢给爬虫中间件处理, 处理结果将将给self.call_spider处理
            return await self.spidermw.scrape_response(self.call_spider, result, request, spider)
        else:
            try:
                # 处理下载错误或经过下载中间件时出现的错误
                return await self.call_spider(result, request, spider)
            except (Exception, BaseException) as e:
                await self._log_download_errors(e, result, request, spider)

    async def call_spider(self, result, request, spider):
        if isinstance(result, Response):
            # 将Response丢给spider的解析函数
            callback = request.callback or spider._parse
            warn_on_generator_with_return_value(spider, callback)
            result.request = request
            # 将parse解析出的结果,变成可迭代对象
            return await call_helper(callback, result, **result.request.cb_kwargs)
        else:
            if request.errback is None:
                raise result
            # 下载中间件或下载结果出现错误,回调request中的errback函数
            warn_on_generator_with_return_value(spider, request.errback)
            return await call_helper(request.errback, result)

    async def handle_spider_error(self, exc, request, response, spider):
        if isinstance(exc, CloseSpider):
            asyncio.create_task(self.crawler.engine.close_spider(spider, exc.reason or 'cancelled'))
            return
        logkws = self.logformatter.spider_error(exc, request, response, spider)
        logger.log(
            *logformatter_adapter(logkws),
            exc_info=exc,
            extra={'spider': spider}
        )
        await self.signals.send_catch_log(
            signal=signals.spider_error,
            failure=exc, response=response,
            spider=spider
        )
        self.crawler.stats.inc_value(
            "spider_exceptions/%s" % exc.__class__.__name__,
            spider=spider
        )

    async def handle_spider_output(self, result, request, response, spider):
        if not result:
            return
        try:
            async for elem in result:
                asyncio.create_task(self._process_spidermw_output(elem, request, response, spider))
        except (Exception, BaseException) as e:
            await self.handle_spider_error(e, request, response, spider)

    async def _process_spidermw_output(self, output, request, response, spider):
        """Process each Request/Item (given in the output parameter) returned
        from the given spider
        """
        async with self.concurrent_items_semaphore:
            if isinstance(output, Request):
                await self.crawler.engine.crawl(request=output, spider=spider)
            elif is_item(output):
                self.slot.itemproc_size += 1
                item = await self.itemproc.process_item(output, spider)
                process_item_method = getattr(spider, 'process_item', None)
                if process_item_method:
                    await call_helper(process_item_method, item)
                await self._itemproc_finished(output, item, response, spider)
            elif output is None:
                pass
            else:
                typename = type(output).__name__
                logger.error(
                    'Spider must return request, item, or None, got %(typename)r in %(request)s',
                    {'request': request, 'typename': typename},
                    extra={'spider': spider},
                )

    async def _log_download_errors(self, spider_exception, download_exception, request, spider):
        """
        处理并记录错误
        :param spider_exception: 将download_exception丢给request.errback处理时,触发的新错误
        :param download_exception: 下载过程发生的错误,或中间件中发生的错误
        :param request:
        :param spider:
        :return:
        """
        if isinstance(download_exception, (Exception, BaseException)) \
                and not isinstance(download_exception, IgnoreRequest):
            logkws = self.logformatter.download_error(download_exception, request, spider)
            logger.log(
                *logformatter_adapter(logkws),
                extra={'spider': spider},
                exc_info=download_exception,
            )

        if spider_exception is not download_exception:
            raise spider_exception

    async def _itemproc_finished(self, output, item, response, spider):
        """ItemProcessor finished for the given ``item`` and returned ``output``
        """
        self.slot.itemproc_size -= 1
        if isinstance(output, (Exception, BaseException)):
            if isinstance(output, DropItem):
                logkws = self.logformatter.dropped(item, output, response, spider)
                if logkws is not None:
                    logger.log(*logformatter_adapter(logkws), extra={'spider': spider})
                return await self.signals.send_catch_log_deferred(
                    signal=signals.item_dropped, item=item, response=response,
                    spider=spider, exception=output)
            else:
                logkws = self.logformatter.item_error(item, output, response, spider)
                logger.log(*logformatter_adapter(logkws), extra={'spider': spider},
                           exc_info=output)
                return await self.signals.send_catch_log_deferred(
                    signal=signals.item_error, item=item, response=response,
                    spider=spider, failure=output)
        else:
            logkws = self.logformatter.scraped(output, response, spider)
            if logkws is not None:
                logger.log(*logformatter_adapter(logkws), extra={'spider': spider})
            return await self.signals.send_catch_log_deferred(
                signal=signals.item_scraped, item=output, response=response,
                spider=spider)



from .middleware_downloader import DownloaderMiddlewareManager
from .middleware_Itempipeline import ItemPipelineManager
from .middleware_spider import SpiderMiddlewareManager
from .middleware_extension import ExtensionManager

__all__ = [
    "DownloaderMiddlewareManager", "ItemPipelineManager",
    "SpiderMiddlewareManager", "ExtensionManager"
]



import asyncio
import random
from collections import deque
from datetime import datetime
from time import time

from scrapy import signals, Request
from aioscrapy.core.downloader.handlers import DownloadHandlers
from aioscrapy.middleware import DownloaderMiddlewareManager
from scrapy.http import Response
from scrapy.resolver import dnscache
from scrapy.utils.httpobj import urlparse_cached


class Slot:
    """Downloader slot"""

    def __init__(self, concurrency, delay, randomize_delay):
        self.concurrency = concurrency
        self.delay = delay
        self.randomize_delay = randomize_delay

        self.active = set()
        self.queue = deque()
        self.transferring = set()
        self.lastseen = 0
        self.delay_run = False

    def free_transfer_slots(self):
        return self.concurrency - len(self.transferring)

    def download_delay(self):
        if self.randomize_delay:
            return random.uniform(0.5 * self.delay, 1.5 * self.delay)
        return self.delay

    def close(self):
        self.delay_run = True

    def __repr__(self):
        cls_name = self.__class__.__name__
        return "%s(concurrency=%r, delay=%0.2f, randomize_delay=%r)" % (
            cls_name, self.concurrency, self.delay, self.randomize_delay)

    def __str__(self):
        return (
                "<downloader.Slot concurrency=%r delay=%0.2f randomize_delay=%r "
                "len(active)=%d len(queue)=%d len(transferring)=%d lastseen=%s>" % (
                    self.concurrency, self.delay, self.randomize_delay,
                    len(self.active), len(self.queue), len(self.transferring),
                    datetime.fromtimestamp(self.lastseen).isoformat()
                )
        )


def _get_concurrency_delay(concurrency, spider, settings):
    delay = settings.getfloat('DOWNLOAD_DELAY')
    if hasattr(spider, 'download_delay'):
        delay = spider.download_delay

    if hasattr(spider, 'max_concurrent_requests'):
        concurrency = spider.max_concurrent_requests

    return concurrency, delay


class Downloader:
    DOWNLOAD_SLOT = 'download_slot'

    def __init__(self, crawler):
        self.settings = crawler.settings
        self.signals = crawler.signals
        self.slots = {}
        self.active = set()
        self.handlers = DownloadHandlers(crawler)
        self.total_concurrency = self.settings.getint('CONCURRENT_REQUESTS')
        self.domain_concurrency = self.settings.getint('CONCURRENT_REQUESTS_PER_DOMAIN')
        self.ip_concurrency = self.settings.getint('CONCURRENT_REQUESTS_PER_IP')
        self.randomize_delay = self.settings.getbool('RANDOMIZE_DOWNLOAD_DELAY')
        self.middleware = DownloaderMiddlewareManager.from_crawler(crawler)
        self._slot_gc_loop = True
        asyncio.create_task(self._slot_gc(60))

    async def fetch(self, request, spider, _handle_downloader_output):
        self.active.add(request)
        key, slot = self._get_slot(request, spider)
        request.meta[self.DOWNLOAD_SLOT] = key

        slot.active.add(request)
        await self.signals.send_catch_log(signal=signals.request_reached_downloader,
                                          request=request,
                                          spider=spider)
        slot.queue.append((request, _handle_downloader_output))
        asyncio.create_task(self._process_queue(spider, slot))

    async def _process_queue(self, spider, slot):
        if slot.delay_run:
            return

        # Delay queue processing if a download_delay is configured
        now = time()
        delay = slot.download_delay()
        if delay:
            penalty = delay - now + slot.lastseen
            if penalty > 0:
                slot.delay_run = True
                await asyncio.sleep(penalty)
                slot.delay_run = False
                asyncio.create_task(self._process_queue(spider, slot))
                return

        # Process enqueued requests if there are free slots to transfer for this slot
        while slot.queue and slot.free_transfer_slots() > 0:
            slot.lastseen = now
            request, _handle_downloader_output = slot.queue.popleft()
            asyncio.create_task(self._download(slot, request, spider, _handle_downloader_output))
            # prevent burst if inter-request delays were configured
            if delay:
                asyncio.create_task(self._process_queue(spider, slot))
                break

    async def _download(self, slot, request, spider, _handle_downloader_output):
        slot.transferring.add(request)
        response = None
        try:
            response = await self.middleware.process_request(spider, request)
            if response is None or isinstance(response, Request):
                request = response or request
                response = await self.handlers.download_request(request, spider)
        except (Exception, BaseException) as exc:
            response = await self.middleware.process_exception(spider, request, exc)
        else:
            try:
                response = await self.middleware.process_response(spider, request, response)
            except (Exception, BaseException) as exc:
                response = exc
        finally:
            slot.transferring.remove(request)
            slot.active.remove(request)
            self.active.remove(request)
            await self.signals.send_catch_log(signal=signals.request_left_downloader,
                                              request=request,
                                              spider=spider)
            asyncio.create_task(self._process_queue(spider, slot))
            if isinstance(response, Response):
                response.request = request
                await self.signals.send_catch_log(signal=signals.response_downloaded,
                                                  response=response,
                                                  request=request,
                                                  spider=spider)
            asyncio.create_task(_handle_downloader_output(response, request, spider))

    def close(self):
        self._slot_gc_loop = False
        for slot in self.slots.values():
            slot.close()

    async def _slot_gc(self, age=60):
        mintime = time() - age
        for key, slot in list(self.slots.items()):
            if not slot.active and slot.lastseen + slot.delay < mintime:
                self.slots.pop(key).close()
        await asyncio.sleep(age)
        if self._slot_gc_loop:
            asyncio.create_task(self._slot_gc())

    def needs_backout(self):
        return len(self.active) >= self.total_concurrency

    def _get_slot(self, request, spider):
        key = self._get_slot_key(request, spider)
        if key not in self.slots:
            conc = self.ip_concurrency if self.ip_concurrency else self.domain_concurrency
            conc, delay = _get_concurrency_delay(conc, spider, self.settings)
            self.slots[key] = Slot(conc, delay, self.randomize_delay)

        return key, self.slots[key]

    def _get_slot_key(self, request, spider):
        if self.DOWNLOAD_SLOT in request.meta:
            return request.meta[self.DOWNLOAD_SLOT]

        key = urlparse_cached(request).hostname or ''
        if self.ip_concurrency:
            key = dnscache.get(key, key)

        return key



import pickle
import json

__all__ = ['PickleCompat', 'JsonCompat']


def _request_body_byte2str(obj):
    obj['body'] = obj['body'].decode(obj['_encoding'])
    return obj


def _request_body_str2byte(obj):
    obj['body'] = obj['body'].encode(obj['_encoding'])
    return obj


class PickleCompat:
    @staticmethod
    def loads(s):
        return pickle.loads(s)

    @staticmethod
    def dumps(obj):
        return pickle.dumps(obj, protocol=-1)


class JsonCompat:
    @staticmethod
    def loads(s):
        return _request_body_str2byte(json.loads(s))

    @staticmethod
    def dumps(obj):
        print(obj)
        return json.dumps(_request_body_byte2str(obj))



# _*_ coding: utf-8 _*_

import asyncio
import logging
from time import time

from scrapy import signals
from scrapy.exceptions import DontCloseSpider
from scrapy.http import Response
from scrapy.http.request import Request
from scrapy.utils.log import (
    logformatter_adapter)
from scrapy.utils.misc import load_object

from aioscrapy.utils.tools import call_helper
from .scraper import Scraper

logger = logging.getLogger(__name__)


class Slot:

    def __init__(self, start_requests, close_if_idle, scheduler):
        self.closing = None
        self.inprogress = set()  # requests in progress
        self.start_requests = iter(start_requests)
        self.close_if_idle = close_if_idle
        self.scheduler = scheduler
        self.heartbeat = None

    def add_request(self, request):
        self.inprogress.add(request)

    def remove_request(self, request):
        self.inprogress.remove(request)
        self._maybe_fire_closing()

    async def close(self):
        self.closing = asyncio.Future()
        self._maybe_fire_closing()
        await self.closing

    def _maybe_fire_closing(self):
        if self.closing and not self.inprogress:
            if self.heartbeat:
                self.heartbeat.cancel()
            self.closing.set_result(None)


class ExecutionEngine(object):

    def __init__(self, crawler, spider_closed_callback):
        self.start_time = time()
        self.crawler = crawler
        self.settings = crawler.settings
        self.signals = crawler.signals
        self.logformatter = crawler.logformatter
        self.slot = None
        self.spider = None
        self.scheduler = None
        self.running = False
        self.paused = False
        self.scheduler_cls = load_object(self.settings['SCHEDULER'])
        downloader_cls = load_object(self.settings['DOWNLOADER'])
        self.downloader = downloader_cls(crawler)
        self.scraper = Scraper(crawler)
        self._spider_closed_callback = spider_closed_callback

    async def start(self, spider, start_requests=()):
        """Start the execution engine"""
        if self.running:
            raise RuntimeError("Engine already running")
        self.start_time = time()
        await self.signals.send_catch_log_deferred(signal=signals.engine_started)
        self.running = True
        self._closewait = asyncio.Future()
        await self.open_spider(spider, start_requests, close_if_idle=True)
        await self._closewait

    async def stop(self):
        """Stop the execution engine gracefully"""
        if not self.running:
            raise RuntimeError("Engine not running")
        self.running = False
        await self._close_all_spiders()
        await self.signals.send_catch_log_deferred(signal=signals.engine_stopped)
        self._closewait.set_result(None)

    async def close(self):
        """Close the execution engine gracefully.

        If it has already been started, stop it. In all cases, close all spiders
        and the downloader.
        """
        if self.running:
            # Will also close spiders and downloader
            await self.stop()
        elif self.open_spiders:
            # Will also close downloader
            await self._close_all_spiders()
        else:
            self.downloader.close()

    def pause(self):
        """Pause the execution engine"""
        self.paused = True

    def unpause(self):
        """Resume the execution engine"""
        self.paused = False

    async def _next_request(self, spider):
        slot = self.slot
        if not slot:
            return

        if self.paused:
            return

        while not self._needs_backout(spider):
            request = await call_helper(slot.scheduler.next_request)
            if not request:
                break
            slot.add_request(request)
            asyncio.create_task(self.downloader.fetch(request, spider, self._handle_downloader_output))

        if slot.start_requests and not self._needs_backout(spider):
            try:
                request = next(slot.start_requests)
            except StopIteration:
                slot.start_requests = None
            except Exception:
                slot.start_requests = None
                logger.error('Error while obtaining start requests',
                             exc_info=True, extra={'spider': spider})
            else:
                await self.crawl(request, spider)

        if await self.spider_is_idle(spider) and slot.close_if_idle:
            await self._spider_idle(spider)

    def _needs_backout(self, spider):
        return (
                not self.running
                or self.slot.closing
                or self.downloader.needs_backout()
                or self.scraper.slot.needs_backout()
        )

    async def _handle_downloader_output(self, result, request, spider):
        try:
            if not isinstance(result, (Request, Response, Exception, BaseException)):
                raise TypeError(
                    "Incorrect type: expected Request, Response or Failure, got %s: %r"
                    % (type(result), result)
                )

            if isinstance(result, Request):
                await self.crawl(result, spider)
                return

            if isinstance(result, Response):
                result.request = request
                logkws = self.logformatter.crawled(request, result, spider)
                if logkws is not None:
                    logger.log(*logformatter_adapter(logkws), extra={'spider': spider})
                await self.signals.send_catch_log(signals.response_received,
                                                  response=result, request=request, spider=spider)

        finally:
            self.slot.remove_request(request)
        await self.scraper.enqueue_scrape(result, request, spider)
        asyncio.create_task(self._next_request(self.spider))

    async def spider_is_idle(self, spider):
        if not self.scraper.slot.is_idle():
            # scraper is not idle
            return False

        if self.downloader.active:
            # downloader has pending requests
            return False

        if self.slot.start_requests is not None:
            # not all start requests are handled
            return False

        if self.slot.inprogress:
            # not all start requests are handled
            return False

        if await call_helper(self.slot.scheduler.has_pending_requests):
            # scheduler has pending requests
            return False

        return True

    @property
    def open_spiders(self):
        return (self.spider,) if self.spider else set()

    def has_capacity(self):
        """Does the engine have capacity to handle more spiders"""
        return not bool(self.slot)

    async def crawl(self, request, spider):  # 将网址 请求加入队列
        if spider not in self.open_spiders:
            raise RuntimeError("Spider %r not opened when crawling: %s" % (spider.name, request))

        await self.signals.send_catch_log(signals.request_scheduled, request=request, spider=spider)
        if asyncio.iscoroutinefunction(self.slot.scheduler.enqueue_request):
            succ = await self.slot.scheduler.enqueue_request(request)
        else:
            succ = self.slot.scheduler.enqueue_request(request)
        if not succ:
            await self.signals.send_catch_log(signals.request_dropped, request=request, spider=spider)
        asyncio.create_task(self._next_request(spider))

    async def open_spider(self, spider, start_requests=(), close_if_idle=True):
        if not self.has_capacity():
            raise RuntimeError("No free spider slot when opening %r" % spider.name)
        logger.info("Spider opened", extra={'spider': spider})
        scheduler = await call_helper(self.scheduler_cls.from_crawler, self.crawler)
        start_requests = await call_helper(self.scraper.spidermw.process_start_requests, start_requests, spider)
        slot = Slot(start_requests, close_if_idle, scheduler)
        self.slot = slot
        self.spider = spider
        await call_helper(scheduler.open, spider)
        await call_helper(self.scraper.open_spider, spider)
        await call_helper(self.crawler.stats.open_spider, spider)
        await self.signals.send_catch_log_deferred(signals.spider_opened, spider=spider)
        slot.heartbeat = asyncio.create_task(self.heart_beat(5, spider, slot))
        asyncio.create_task(self._next_request(spider))

    async def _close_all_spiders(self):
        dfds = [self.close_spider(s, reason='shutdown') for s in self.open_spiders]
        await asyncio.gather(*dfds)

    async def close_spider(self, spider, reason='cancelled'):
        """Close (cancel) spider and clear all its outstanding requests"""

        slot = self.slot
        if slot.closing:
            return slot.closing
        logger.info("Closing spider (%(reason)s)",
                    {'reason': reason},
                    extra={'spider': spider})

        await slot.close()

        async def close_handler(callback, *args, errmsg='', **kwargs):
            try:
                if asyncio.iscoroutinefunction(callback):
                    await callback(*args, **kwargs)
                else:
                    callback(*args, **kwargs)
            except (Exception, BaseException) as e:
                logger.error(
                    errmsg,
                    exc_info=e,
                    extra={'spider': spider}
                )

        await close_handler(self.downloader.close, errmsg='Downloader close failure')

        await close_handler(self.scraper.close_spider, spider, errmsg='Scraper close failure')

        await close_handler(self.slot.scheduler.close, reason, errmsg='Scheduler close failure')

        await close_handler(self.signals.send_catch_log_deferred, signal=signals.spider_closed, spider=spider,
                            reason=reason, errmsg='Error while sending spider_close signal')

        await close_handler(self.crawler.stats.close_spider, spider, reason=reason, errmsg='Stats close failure')

        logger.info("Spider closed (%(reason)s)", {'reason': reason}, extra={'spider': spider})

        await close_handler(setattr, self, 'slot', None, errmsg='Error while unassigning slot')

        await close_handler(setattr, self, 'spider', None, errmsg='Error while unassigning spider')

        await self._spider_closed_callback()

    async def _spider_idle(self, spider):
        """Called when a spider gets idle. This function is called when there
        are no remaining pages to download or schedule. It can be called
        multiple times. If some extension raises a DontCloseSpider exception
        (in the spider_idle signal handler) the spider is not closed until the
        next loop and this function is guaranteed to be called (at least) once
        again for this spider.
        """
        res = await self.signals.send_catch_log(signals.spider_idle, spider=spider, dont_log=DontCloseSpider)
        if any(isinstance(x, DontCloseSpider) for _, x in res):
            return

        if await self.spider_is_idle(spider):
            await self.close_spider(spider, reason='finished')

    async def heart_beat(self, delay, spider, slot):
        while not slot.closing:
            await asyncio.sleep(delay)
            asyncio.create_task(self._next_request(spider))




try:
    from asyncio.exceptions import TimeoutError
except:
    from concurrent.futures._base import TimeoutError
from aiohttp.client_exceptions import ClientError

from scrapy.downloadermiddlewares.retry import RetryMiddleware as ScrapyRetryMiddleware


class RetryMiddleware(ScrapyRetryMiddleware):

    EXCEPTIONS_TO_RETRY = ScrapyRetryMiddleware.EXCEPTIONS_TO_RETRY + (TimeoutError, ClientError)



from typing import Union, Tuple

from aioredis import BlockingConnectionPool, Redis


class AioRedisManager(object):
    _clients = {}

    @staticmethod
    def parse_params(alias_or_params:  Union[str, dict]) -> Tuple[str, Union[dict, None]]:
        """
        将参数中的别名和redis参数提取出来
        """
        if isinstance(alias_or_params, str):
            return alias_or_params, None

        redis_params = alias_or_params.copy()
        url = redis_params.get('url')
        db = redis_params.get('db', '')
        alias = redis_params.pop('alias', url + str(db))
        return alias, redis_params

    def create(self, params: Union[dict], alias=None) -> Redis:
        if alias is None:
            alias, params = self.parse_params(params)
        url = params.pop('url')
        redis = Redis(connection_pool=BlockingConnectionPool.from_url(url, **params))
        return self._clients.setdefault(alias, redis)

    def get(self, alias_or_params: Union[str, dict]) -> Redis:
        """获取redis链接"""
        assert isinstance(alias_or_params, (str, dict)), "alias_or_params 参数不正确"
        alias, redis_params = self.parse_params(alias_or_params)
        redis = self._clients.get(alias)
        if redis:
            return redis
        return self.create(redis_params, alias)

    async def close(self, alias_or_params: Union[str, dict]):
        """关闭指定redis pool"""
        assert isinstance(alias_or_params, (str, dict)), "alias_or_params 参数不正确"
        alias, _ = self.parse_params(alias_or_params)
        redis = self._clients.pop(alias, None)
        if redis:
            await redis.close()
            await redis.connection_pool.disconnect()

    async def close_all(self):
        for alias in list(self._clients.keys()):
            await self.close(alias)

    async def from_settings(self, settings: "scrapy.settings.Setting"):
        redis_args = settings.getdict('REDIS_ARGS')
        return self.get(redis_args)

    async def from_crawler(self, crawler):
        return await self.from_settings(crawler.settings)


redis_manager = AioRedisManager()


if __name__ == '__main__':
    import asyncio

    async def test():
        redis = redis_manager.get({
                'alias': 'xx',
                'url': 'redis://:erpteam_redis@192.168.5.216:6381/9',
                'max_connections': 4
            })

        await redis.zadd('key1', {'value': 1})

        async with redis.pipeline(transaction=True) as pipe:
            results, count = await (
                pipe.zrange('key1', 0, 0)
                    .zremrangebyrank('key1', 0, 0)
                    .execute()
            )

        print(results)
        await redis_manager.close_all()


    # asyncio.run(test(), debug=True)
    asyncio.get_event_loop().run_until_complete(test())




import logging
from asyncio import iscoroutinefunction

from scrapy.exceptions import _InvalidOutput
from scrapy.http import Request, Response
from scrapy.utils.conf import build_component_list

from .middleware import MiddlewareManager

logger = logging.getLogger(__name__)


class DownloaderMiddlewareManager(MiddlewareManager):
    component_name = 'downloader middleware'

    @classmethod
    def _get_mwlist_from_settings(cls, settings):
        return build_component_list(
            settings.getwithbase('DOWNLOADER_MIDDLEWARES'))

    def _add_middleware(self, mw):
        if hasattr(mw, 'process_request'):
            self.methods['process_request'].append(mw.process_request)
        if hasattr(mw, 'process_response'):
            self.methods['process_response'].appendleft(mw.process_response)
        if hasattr(mw, 'process_exception'):
            self.methods['process_exception'].appendleft(mw.process_exception)

    async def process_request(self, spider, request):
        for method in self.methods['process_request']:
            if iscoroutinefunction(method):
                response = await method(request=request, spider=spider)
            else:
                response = method(request=request, spider=spider)
            if response is not None and not isinstance(response, (Response, Request)):
                raise _InvalidOutput(
                    "Middleware %s.process_request must return None, Response or Request, got %s"
                    % (method.__self__.__class__.__name__, response.__class__.__name__)
                )
            if response:
                return response

    async def process_response(self, spider, request, response):
        if response is None:
            raise TypeError("Received None in process_response")
        elif isinstance(response, Request):
            return response

        for method in self.methods['process_response']:
            if iscoroutinefunction(method):
                response = await method(request=request, response=response, spider=spider)
            else:
                response = method(request=request, response=response, spider=spider)
            if not isinstance(response, (Response, Request)):
                raise _InvalidOutput(
                    "Middleware %s.process_response must return Response or Request, got %s"
                    % (method.__self__.__class__.__name__, type(response))
                )
            if isinstance(response, Request):
                return response
        return response

    async def process_exception(self, spider, request, exception):
        for method in self.methods['process_exception']:
            if iscoroutinefunction(method):
                response = await method(request=request, exception=exception, spider=spider)
            else:
                response = method(request=request, exception=exception, spider=spider)
            if response is not None and not isinstance(response, (Response, Request)):
                raise _InvalidOutput(
                    "Middleware %s.process_exception must return None, Response or Request, got %s"
                    % (method.__self__.__class__.__name__, type(response))
                )
            if response:
                return response
        return exception




from scrapy.settings import Settings
from scrapy.settings import BaseSettings
from aioscrapy.settings import aio_settings


class AioSettings(Settings):
    
    def __init__(self, values=None, priority='project'):
        super().__init__()
        self.setmodule(aio_settings, 'default')

        for name, val in self.items():
            if isinstance(val, dict):
                self.set(name, BaseSettings(val, 'default'), 'default')
        self.update(values, priority)

    def getwithbase(self, name):
        """Get a composition of a dictionary-like setting and its `_BASE`
        counterpart.

        :param name: name of the dictionary-like setting
        :type name: str
        """
        compbs = BaseSettings()
        compbs.update(self[name + '_BASE'])
        compbs.update(self[name])
        compbs.update(self[name + '_USER'])
        return compbs



from typing import Union, Tuple
from contextlib import asynccontextmanager

from aiomysql import create_pool


class AioMysqlManager(object):
    _clients = {}

    @staticmethod
    def parse_params(alias_or_params: Union[str, dict]) -> Tuple[str, Union[dict, None]]:
        if isinstance(alias_or_params, str):
            return alias_or_params, None

        mysql_params = alias_or_params.copy()
        alias = mysql_params.pop('alias', f"{mysql_params['host']}{mysql_params['port']}")
        return alias, mysql_params

    async def create(self, params: Union[dict], alias=None):
        if alias is None:
            alias, params = self.parse_params(params)
        mysql_pool = await create_pool(**params)
        return self._clients.setdefault(alias, mysql_pool)

    @asynccontextmanager
    async def get(self, alias_or_params: Union[str, dict], ping=False):
        """获取数据库链接和数据库游标"""
        assert isinstance(alias_or_params, (str, dict)), "alias_or_params 参数不正确"
        alias, params = self.parse_params(alias_or_params)
        mysql_pool = self._clients.get(alias)
        if not mysql_pool:
            mysql_pool = await self.create(params, alias)

        conn = await mysql_pool.acquire()
        if ping:
            await conn.ping()
        cur = await conn.cursor()
        yield conn, cur
        await cur.close()
        await mysql_pool.release(conn)

    async def close(self, alias_or_params: Union[str, dict]):
        assert isinstance(alias_or_params, (str, dict)), "alias_or_params 参数不正确"
        alias, _ = self.parse_params(alias_or_params)
        mysql_pool = self._clients.get(alias)
        if mysql_pool:
            mysql_pool.close()
            await mysql_pool.wait_closed()

    async def close_all(self):
        for alias in list(self._clients.keys()):
            await self.close(alias)


mysql_manager = AioMysqlManager()

if __name__ == '__main__':
    import asyncio


    async def test():
        mysql_pool = await mysql_manager.create({
            'alias': 'xx',
            'db': 'test',
            'user': 'root',
            'password': 'root',
            'host': '192.168.5.237',
            'port': 3306,
            'charset': 'utf8',
        })

        # 方式一:
        try:
            conn = await mysql_pool.acquire()
            cur = await conn.cursor()
            print(await cur.execute('select 1'))
            # await conn.commit()
        finally:
            await cur.close()
            await mysql_pool.release(conn)

        # 方式二:
        async with mysql_manager.get('xx') as (conn, cur):
            print(await cur.execute('select 1'))
            # await conn.commit()

        await mysql_manager.close_all()


    asyncio.run(test())



import os

from setuptools import setup


def list_dir(dir_path):
    result = [dir_path]
    for file in os.listdir(dir_path):
        if os.path.isdir(os.path.join(dir_path, file)):
            result.extend(list_dir(os.path.join(dir_path, file)))
    return result


NAME = "aio_scrapy"
PACKAGES = list_dir('aioscrapy')
DESCRIPTION = "Replace twisted of Scrapy with asyncio and aiohttp"
LONG_DESCRIPTION = ''
URL = "https://github.com/conlin-huang/aioscrapy.git"
AUTHOR = "conlin"
AUTHOR_EMAIL = "995018884@qq.com"
VERSION = "0.0.1"
LICENSE = "MIT"

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
    ],
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    packages=PACKAGES,
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        "scrapy-redis",
        "scrapy",
        "aiohttp",
        "aioredis >= 2.0.0",
        "aiomysql"
    ],
    keywords=[
        'aio-scrapy', 'scrapy', 'aioscrapy',
        'scrapy redis', 'asyncio', 'spider',
    ]
)



import asyncio
import logging
import ssl
import sys

from scrapy.http import Headers
from aioscrapy.https import TextResponse
import aiohttp

logger = logging.getLogger(__name__)


class AioHttpDownloadHandler:
    session = None

    def __init__(self, settings, crawler=None):
        self._crawler = crawler
        self.settings = settings

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings, crawler)

    def get_session(self, *arg, **kw):
        if self.session is None:
            if sys.platform.startswith('win'):
                asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
            self.session = aiohttp.ClientSession(*arg, **kw)
        return self.session

    async def download_request(self, request, spider):
        kwargs = {}

        if request.headers:
            headers = request.headers
        elif self.settings.get('headers', False):
            headers = self.settings.get('headers')
        else:
            headers = {}
        if isinstance(headers, Headers):
            headers = headers.to_unicode_dict()
        kwargs['headers'] = headers

        timeout = self.settings.get("DOWNLOAD_TIMEOUT", None)
        if timeout:
            kwargs['timeout'] = timeout

        proxy = request.meta.get("proxy")
        if proxy:
            kwargs["proxy"] = proxy
            logger.info(f"使用代理{proxy}抓取: {request.url}")

        cookies = request.cookies
        if cookies:
            kwargs['cookies'] = dict(cookies)

        ssl_ciphers = request.meta.get('TLS_CIPHERS')
        if ssl_ciphers:
            context = ssl.create_default_context()
            context.set_ciphers(ssl_ciphers)
            kwargs['ssl'] = context

        session = self.get_session()
        method = getattr(session, request.method.lower())
        if request.body:
            kwargs['data'] = request.body
        response = await method(request.url, **kwargs)
        content = await response.read()
        return TextResponse(str(response.url),
                            status=response.status,
                            headers=response.headers,
                            body=content,
                            cookies=response.cookies)

    async def close(self):
        if self.session is not None:
            await self.session.close()

        # Wait 250 ms for the underlying SSL connections to close
        # https://docs.aiohttp.org/en/latest/client_advanced.html#graceful-shutdown
        await asyncio.sleep(0.250)



# _*_ coding: utf-8 _*_

import asyncio
from types import CoroutineType, GeneratorType, AsyncGeneratorType


async def call_helper(f, *args, **kwargs):
    if asyncio.iscoroutinefunction(f):
        return await f(*args, **kwargs)
    return f(*args, **kwargs)


async def async_generator_wrapper(wrapped):
    """将传入的对象变成AsyncGeneratorType类型"""
    if isinstance(wrapped, AsyncGeneratorType):
        return wrapped

    elif isinstance(wrapped, CoroutineType):
        async def anonymous(c):
            yield await c
        return anonymous(wrapped)

    elif isinstance(wrapped, GeneratorType):
        async def anonymous(c):
            for r in c:
                yield r
        return anonymous(wrapped)

    else:
        async def anonymous(c):
            yield c
        return anonymous(wrapped)


def get_project_settings():
    import os
    import pickle
    import warnings

    from scrapy.utils.conf import init_env
    from aioscrapy.settings import AioSettings
    from scrapy.exceptions import ScrapyDeprecationWarning

    ENVVAR = 'SCRAPY_SETTINGS_MODULE'

    if ENVVAR not in os.environ:
        project = os.environ.get('SCRAPY_PROJECT', 'default')
        init_env(project)
    settings = AioSettings()
    settings_module_path = os.environ.get(ENVVAR)
    if settings_module_path:
        settings.setmodule(settings_module_path, priority='project')

    pickled_settings = os.environ.get("SCRAPY_PICKLED_SETTINGS_TO_OVERRIDE")
    if pickled_settings:
        warnings.warn("Use of environment variable "
                      "'SCRAPY_PICKLED_SETTINGS_TO_OVERRIDE' "
                      "is deprecated.", ScrapyDeprecationWarning)
        settings.setdict(pickle.loads(pickled_settings), priority='project')

    scrapy_envvars = {k[7:]: v for k, v in os.environ.items() if
                      k.startswith('SCRAPY_')}
    valid_envvars = {
        'CHECK',
        'PICKLED_SETTINGS_TO_OVERRIDE',
        'PROJECT',
        'PYTHON_SHELL',
        'SETTINGS_MODULE',
    }
    setting_envvars = {k for k in scrapy_envvars if k not in valid_envvars}
    if setting_envvars:
        setting_envvar_list = ', '.join(sorted(setting_envvars))
        warnings.warn(
            'Use of environment variables prefixed with SCRAPY_ to override '
            'settings is deprecated. The following environment variables are '
            f'currently defined: {setting_envvar_list}',
            ScrapyDeprecationWarning
        )
    settings.setdict(scrapy_envvars, priority='project')

    return settings


def singleton(cls):
    _instance = {}

    def _singleton(*args, **kwargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)
        return _instance[cls]

    return _singleton


def exec_js_func(js_file_path, func_name, func_params=None, cwd_path=None, cmd_path='node'):
    """
    调用node执行js文件
    :param js_file_path: js文件路径
    :param func_name: 要调用的js文件内的函数名称
    :param func_params: func的参数
    :param cwd_path: node_modules文件所在路径，如果不指定将使用全局的node_modules
    :param cmd_path: node的位置， 例如：r'D:\\path\to\node.exe'
    :return: func_name函数的执行结果
    """
    import execjs

    if func_params is None:
        func_params = []
    name = "MyNode"
    execjs.register(name, execjs._external_runtime.ExternalRuntime(
        name="Node.js (V8)",
        command=[cmd_path],
        encoding='UTF-8',
        runner_source=execjs._runner_sources.Node
    ))
    with open(js_file_path, 'r', encoding='utf-8') as f:
        lines = f.readlines()
    js = ''.join(lines)
    js_context = execjs.get(name).compile(js, cwd=cwd_path)
    return js_context.call(func_name, *func_params)
